/*
 * ti_usb - a class used to handle USB comms for ti calculators
 *
 * Copyright (C) 2010 Michael Jordan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "ti_usb.h"


	ti_usb::ti_usb() {
		ti_usb::is_inited = false;
		if(libusb_init(&(ti_usb::ctx)) != 0) return;
		libusb_set_debug(ti_usb::ctx, 3);
		if(find_calc() != 0) {
			libusb_exit(ti_usb::ctx);
			ti_usb::ctx = NULL;
			return;
		}
		if(unload_kern_driver() != 0) {
			libusb_close(ti_usb::hdl);
			ti_usb::hdl = NULL;
			libusb_exit(ti_usb::ctx);
			ti_usb::ctx = NULL;
			return;
		}
		if(libusb_claim_interface(ti_usb::hdl, 0) != 0) {
			libusb_close(ti_usb::hdl);
			ti_usb::hdl = NULL;
			libusb_exit(ti_usb::ctx);
			ti_usb::ctx = NULL;
			return;
		}
		ti_usb::is_inited = true;
		if(get_id_string() == "") {
			libusb_release_interface(ti_usb::hdl, 0);
			libusb_close(ti_usb::hdl);
			ti_usb::hdl = NULL;
			libusb_exit(ti_usb::ctx);
			ti_usb::ctx = NULL;
			ti_usb::is_inited = false;
		}
		return;
	}


	ti_usb::~ti_usb() {
		if(is_good()) {
			is_inited = false;
			libusb_release_interface(ti_usb::hdl, 0);
			libusb_close(ti_usb::hdl);
			ti_usb::hdl = NULL;
			libusb_exit(ti_usb::ctx);
			ti_usb::ctx = NULL;
		}
	}


	bool ti_usb::is_good() {
		return ti_usb::is_inited;
	}


	bool ti_usb::set_homescreen() {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0037;
			req.data_len = 1;
			req.data = new unsigned char[1];
			req.data[0] = 0x1;
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	bool ti_usb::ls(vector<ti_usb::ls_item_info> &items) {
		if(is_good()) {
			packet pkt_prsd, pkt_formed;
			int ret;

			if(!reset_pipe_long()) return false;
			if(init_chat() != 0) return false;
			pkt_formed.raw_len = 0x1D;
			pkt_formed.raw_type = 0x04;
			pkt_formed.virt_len = 0x17;
			pkt_formed.virt_type = 0x09;
			pkt_formed.hdl = ti_usb::hdl;
			pkt_formed.payload = new unsigned char[0x17];
			pkt_formed.payload[0] = 0x00; pkt_formed.payload[1] = 0x00;
			pkt_formed.payload[2] = 0x00; pkt_formed.payload[3] = 0x06;
			pkt_formed.payload[4] = 0x00; pkt_formed.payload[5] = 0x02;
			pkt_formed.payload[6] = 0x00; pkt_formed.payload[7] = 0x03;
			pkt_formed.payload[8] = 0x00; pkt_formed.payload[9] = 0x05;
			pkt_formed.payload[10] = 0x00; pkt_formed.payload[11] = 0x01;
			pkt_formed.payload[12] = 0x00; pkt_formed.payload[13] = 0x41;
			pkt_formed.payload[14] = 0x00; pkt_formed.payload[15] = 0x42;
			pkt_formed.payload[16] = 0x00; pkt_formed.payload[17] = 0x01;
			pkt_formed.payload[18] = 0x00; pkt_formed.payload[19] = 0x01;
			pkt_formed.payload[20] = 0x00; pkt_formed.payload[21] = 0x01;
			pkt_formed.payload[22] = 0x01;
			if(send_pkt(pkt_formed) != 0) return false;
			delete[] pkt_formed.payload;
			pkt_prsd.hdl = ti_usb::hdl;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return false;
			if(pkt_prsd.raw_type != 5) return false;
			while(pkt_prsd.virt_type != 0xDD00) {
				if(recv_pkt(&pkt_prsd) != 0) return false;
				if(pkt_prsd.virt_type == 0x000A) {
					if(parse_ls_pkt(pkt_prsd.payload, items) != 0) {
						delete[] pkt_prsd.payload;
						return false;
					}
					delete[] pkt_prsd.payload;
				}
			}
			if(!reset_pipe_short()) return false;
			return true;
		}
		return false;
	}


	string ti_usb::get_id_string() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0002;
			req.data_len = 14;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return "";
			}
			string s = "";
			char *buf = new char[req.data_len + 1];
			for(int i = 0; i < req.data_len; i ++)
				buf[i] = req.data[i];
			buf[req.data_len] = 0;
			s.append(buf);
			delete[] req.data;
			delete[] buf;
			return s;
		}
		return "";
	}


	string ti_usb::get_id_list() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0005;
			req.data_len = 512;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return "";
			}
			string s = "";
			char *buf = new char[req.data_len + 1];
			for(int i = 0; i < req.data_len; i ++)
				buf[i] = req.data[i];
			buf[req.data_len] = 0;
			s.append(buf);
			delete[] req.data;
			delete[] buf;
			return s;
		}
		return "";
	}


	int ti_usb::get_free_ram() {
		if(is_good()) {
			param_req req;

			req.param_num = 0x000E;
			req.data_len = 8;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int free_ram = -1;
			free_ram = req.data[7];
			free_ram += (req.data[6] * 0x100);
			free_ram += (req.data[5] * 0x10000);
			free_ram += (req.data[4] * 0x1000000);
			delete[] req.data;
			return free_ram;
		}
		return -1;
	}


	int ti_usb::get_free_flash() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x00011;
			req.data_len = 8;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int free_flash = -1;
			free_flash = req.data[7];
			free_flash += (req.data[6] * 0x100);
			free_flash += (req.data[5] * 0x10000);
			free_flash += (req.data[4] * 0x1000000);
			delete[] req.data;
			return free_flash;
		}
		return -1;
	}


	int ti_usb::get_phys_flash() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0000F;
			req.data_len = 8;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int phys_flash = -1;
			phys_flash = req.data[7];
			phys_flash += (req.data[6] * 0x100);
			phys_flash += (req.data[5] * 0x10000);
			phys_flash += (req.data[4] * 0x1000000);
			delete[] req.data;
			return phys_flash;
		}
		return -1;
	}


	int ti_usb::get_phys_ram() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x000C;
			req.data_len = 8;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int phys_ram = -1;
			phys_ram = req.data[7];
			phys_ram += (req.data[6] * 0x100);
			phys_ram += (req.data[5] * 0x10000);
			phys_ram += (req.data[4] * 0x1000000);
			delete[] req.data;
			return phys_ram;
		}
		return -1;
	}


	int ti_usb::get_prod_num() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0001;
			req.data_len = 4;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int prod_num = -1;
			prod_num = req.data[3];
			prod_num += (req.data[2] * 0x100);
			prod_num += (req.data[1] * 0x10000);
			prod_num += (req.data[0] * 0x1000000);
			delete[] req.data;
			return prod_num;
		}
		return -1;
	}


	int ti_usb::get_hdw_version() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0004;
			req.data_len = 2;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int hdw_version = -1;
			hdw_version = req.data[1];
			hdw_version += (req.data[0] * 0x100);
			delete[] req.data;
			return hdw_version;
		}
		return -1;
	}


	string ti_usb::get_boot_version() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0009;
			req.data_len = 4;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return "";
			}
			stringstream major_hi, major_lo, minor, micro;
			string boot_version = "";
			major_hi.fill('0');
			major_lo.fill('0');
			minor.fill('0');
			micro.fill('0');
			major_hi.width(2);
			major_lo.width(2);
			minor.width(2);
			micro.width(2);
			major_hi << hex << (int)req.data[0];
			major_lo << hex << (int)req.data[1];
			minor << hex << (int)req.data[2];
			micro << hex << (int)req.data[3];
			boot_version = major_hi.str() + major_lo.str() + "." + minor.str() + "." + micro.str();
			delete[] req.data;
			return boot_version;
		}
		return "";
	}


	string ti_usb::get_os_version() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x000B;
			req.data_len = 4;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return "";
			}
			stringstream major_hi, major_lo, minor, micro;
			string boot_version = "";
			major_hi.fill('0');
			major_lo.fill('0');
			minor.fill('0');
			micro.fill('0');
			major_hi.width(2);
			major_lo.width(2);
			minor.width(2);
			micro.width(2);
			major_hi << hex << (int)req.data[0];
			major_lo << hex << (int)req.data[1];
			minor << hex << (int)req.data[2];
			micro << hex << (int)req.data[3];
			boot_version = major_hi.str() + major_lo.str() + "." + minor.str() + "." + micro.str();
			delete[] req.data;
			return boot_version;
		}
		return "";
	}


	int ti_usb::get_dbus_dev_type() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0008;
			req.data_len = 2;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int dbus_dev_type = -1;
			dbus_dev_type = req.data[1];
			dbus_dev_type += (req.data[0] * 0x100);
			delete[] req.data;
			return dbus_dev_type;
		}
		return -1;
	}


	bool ti_usb::is_os_loaded() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x000A;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			bool os_loaded = (req.data[0] == 1);
			delete[] req.data;
			return os_loaded;
		}
		return false;
	}


	int ti_usb::get_lcd_height() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x001F;
			req.data_len = 2;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int lcd_height = -1;
			lcd_height = req.data[1];
			lcd_height += (req.data[0] * 0x100);
			delete[] req.data;
			return lcd_height;
		}
		return -1;
	}


	int ti_usb::get_lcd_width() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x001E;
			req.data_len = 2;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int lcd_width = -1;
			lcd_width = req.data[1];
			lcd_width += (req.data[0] * 0x100);
			delete[] req.data;
			return lcd_width;
		}
		return -1;
	}


	bool ti_usb::disable_clock() {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0024;
			req.data_len = 1;
			req.data = new unsigned char[1];
			req.data[0] = 0;
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	bool ti_usb::enable_clock() {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0024;
			req.data_len = 1;
			req.data = new unsigned char[1];
			req.data[0] = 1;
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	bool ti_usb::is_clock_enabled() {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0024;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			bool ans = (req.data[0] == 1);
			delete[] req.data;
			return ans;
		}
		return false;
	}


	bool ti_usb::set_date_fmt(int fmt) {
		if((is_good()) && (fmt > -1) && (fmt < 9)) {
			param_req req;

			req.param_num = 0x0027;
			req.data_len = 1;
			req.data = new unsigned char[1];
			req.data[0] = (fmt & 0xFF);
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	int ti_usb::get_date_fmt() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0027;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int i = req.data[0];
			delete[] req.data;
			return i;
		}
		return -1;
	}


	bool ti_usb::set_clock_mode(int mode) {
		if((is_good()) && ((mode == 0) || (mode == 1))) {
			param_req req;

			req.param_num = 0x0028;
			req.data_len = 1;
			req.data = new unsigned char[1];
			req.data[0] = mode;
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	int ti_usb::get_clock_mode() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0028;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int clock_mode = req.data[0];
			delete[] req.data;
			return clock_mode;
		}
		return -1;
	}


	bool ti_usb::is_batt_good() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x002D;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			bool batt_good = (req.data[0] == 1);
			delete[] req.data;
			return batt_good;
		}
		return false;
	}


	bool ti_usb::get_homescreen_state() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0037;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			bool is_home = (req.data[0] == 1);
			delete[] req.data;
			return is_home;
		}
		return false;
	}


	bool ti_usb::is_screen_split() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0039;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			bool is_split = (req.data[0] == 1);
			delete[] req.data;
			return is_split;
		}
		return false;
	}


	string ti_usb::get_cert() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0003;
			req.data_len = 18;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return "";
			}
			string s = "";
			char *buf = new char[req.data_len + 1];
			for(int i = 0; i < req.data_len; i ++)
				buf[i] = req.data[i];
			buf[req.data_len] = 0;
			s.append(buf);
			delete[] req.data;
			delete[] buf;
			return s;
		}
		return "";
	}


	int ti_usb::get_lang_id() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0006;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int lang_id = req.data[0];
			delete[] req.data;
			return lang_id;
		}
		return -1;
	}


	int ti_usb::get_sublang_id() {
		if(is_good()) {
			param_req req;
			req.param_num = 0x0007;
			req.data_len = 1;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return -1;
			}
			int sublang_id = req.data[0];
			delete[] req.data;
			return sublang_id;
		}
		return -1;
	}


	bool ti_usb::get_screenshot(string filename) {
		if(is_good()) {
			param_req req;

			req.param_num = 0x22;
			req.data_len = 3840;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			unsigned char *screenshot = new unsigned char[2011];
			screenshot[0] = 0x50; screenshot[1] = 0x34; screenshot[2] = 0x0A;
			screenshot[3] = 0x31; screenshot[4] = 0x36; screenshot[5] = 0x30; screenshot[6] = 0x20;
			screenshot[7] = 0x31; screenshot[8] = 0x30; screenshot[9] = 0x30; screenshot[10] = 0x0A;
			int msglen = 0;
			for(int i = 0; i < 100; i++) {
				for(int j = 0; j < 20; j++) {
					screenshot[msglen + j + 11] = req.data[(i * 30) + j];
				}
				msglen += 20;
			}
			delete[] req.data;
			struct stat st;
			if((stat(filename.c_str(), &st) != 0)) {
				ofstream outfile(filename.c_str(), ios::out | ios::binary);
				outfile.write((char *)screenshot, 2011);
				outfile.close();
			}
			delete[] screenshot;
			return true;
		}
		return false;
	}


	bool ti_usb::set_clock(date_time dt) {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0025;
			req.data_len = 4;
			req.data = new unsigned char[req.data_len];
			unsigned int time_value;
			if(!get_secs_ti(dt, &time_value)) return false;
			req.data[0] = ((time_value & 0xFF000000) / 0x1000000);
			req.data[1] = ((time_value & 0xFF0000) / 0x10000);
			req.data[2] = ((time_value & 0xFF00) / 0x100);
			req.data[3] = (time_value & 0xFF);
			bool ret;
			ret = parameter_set(req);
			delete[] req.data;
			return ret;
		}
		return false;
	}


	bool ti_usb::get_clock(date_time *dt) {
		if(is_good()) {
			param_req req;

			req.param_num = 0x0025;
			req.data_len = 4;
			req.data = new unsigned char[req.data_len];
			if(!parameter_request(&req)) {
				delete[] req.data;
				return false;
			}
			int i = req.data[3];
			i += (req.data[2] * 0x100);
			i += (req.data[1] * 0x10000);
			i += (req.data[0] * 0x1000000);
			delete[] req.data;
			float days = i / 86400;
			float jd = days + 2450449.5;
			int daysecs_h = i - int(days * 86400);
			float tmp2 = (daysecs_h / 3600);
			int daysecs_m = daysecs_h - int(tmp2 * 3600);
			float tmp3 = (daysecs_m / 60);
			int daysecs_s = daysecs_m - int(tmp3 * 60);
			int h = 0;
			int m = 0;
			for(int i = 1; i <= 24; i++) {
				if((i * 3600) > daysecs_h) {
					h = i - 1;
					i = 25;
				}
			}
			for(int i = 1; i <= 60; i++) {
				if((i * 60) > daysecs_m) {
					m = i - 1;
					i = 61;
				}
			}
			dt->hour = h;
			dt->minute = m;
			dt->second = daysecs_s;
			int month;
			int day;
			int year;
			convert_jd_to_gd(jd, &month, &day, &year);
			dt->month = month;
			dt->day = day;
			dt->year = year;
			return true;
		}
		return false;
	}


	bool ti_usb::get_sys_time(date_time *dt) {
		if(is_good()) {
			time_t system_time_t;
			struct tm system_time_tm;

			system_time_t = time(NULL);
			if(system_time_t == -1) return false;
			system_time_tm = *localtime(&system_time_t);
			dt->year = system_time_tm.tm_year + 1900;
			dt->month = system_time_tm.tm_mon + 1;
			dt->day = system_time_tm.tm_mday;
			dt->hour = system_time_tm.tm_hour;
			dt->minute = system_time_tm.tm_min;
			dt->second = system_time_tm.tm_sec;
			return true;
		}
		return false;
	}


	void ti_usb::convert_jd_to_gd(float jd, int *month, int *day, int *year) {
		int x2 = jd - 1721119.5;
		int c2 = (4 * x2 + 3) / 146097;
		int x1 = x2 - int(146097 * c2 / 4);
		int c1 = (100 * x1 + 99) / 36525;
		int x0 = x1 - int(36525 * c1 / 100);
		int j = 100 * c2 + c1;
		int m = int((5 * x0 + 461) / 153);
		int d = x0 - int((153 * m - 457) / 5) + 1;
		if(m > 12) {
			m = m - 12;
			j = j + 1;
		}
		*year = j;
		*month = m;
		*day = d;
		return;
	}


	long double ti_usb::to_julian_days(int Y, int M, int D) {
		int A, B, C, E, F;
		long double JD;

		if((M == 1)||(M == 2)) {
			Y -= 1;
			M += 12;
		}

		A = Y / 100;
		B = A / 4;
		C = (2 - A) + B;
		E = 365.25 * (Y + 4716);
		F = 30.6001 * (M + 1);
		JD = C + D + E + F - 1524.5;

		return JD;
	}


	bool ti_usb::get_secs_ti(date_time des_date, unsigned int *result) {
		long double t1_js, t2_js;

		if((des_date.year < 1997)||(des_date.year > 2084)) return false;

		t1_js = to_julian_days(des_date.year, (des_date.month), des_date.day);
		t2_js = to_julian_days(1997, 1, 1);
		t1_js *= 86400;
		t2_js *= 86400;
		t1_js += ((des_date.hour * 3600) + (des_date.minute * 60) + des_date.second);
		if(t2_js > t1_js) return false;
		*result = (unsigned int)(t1_js - t2_js);

		return true;
	}


	bool ti_usb::bulk_transfer(transfer_info *tinfo) {
		int actual, res;
		bool len_ok;
		res = libusb_bulk_transfer(tinfo->hdl, tinfo->direct, tinfo->data, *tinfo->len, &actual, tinfo->timeout);
		tinfo->actual = actual;
		if(tinfo->direct == out) len_ok = (*tinfo->len == actual);
		else len_ok = true;
		return ((res == 0) && len_ok);
	}


	bool ti_usb::reset_pipe_long() {
		ctl_transfer_info cti;

		cti.hdl = ti_usb::hdl;
		cti.bmRequestType = 0x02;
		cti.bRequest = 0x01;
		cti.wValue = 0x0000;
		cti.wIndex = 0x0081;
		cti.wLength = 0x0000;
		cti.data = NULL;
		cti.timeout = 0;
		if(!control_transfer(&cti)) return false;
		cti.wIndex = 0x0002;
		if(!control_transfer(&cti)) return false;
		sleep(1);
		cti.wIndex = 0x0081;
		if(!control_transfer(&cti)) return false;
		cti.wIndex = 0x0002;
		if(!control_transfer(&cti)) return false;
		sleep(1);
		cti.bmRequestType = 0x80;
		cti.bRequest = 0x06;
		cti.wValue = 0x0100;
		cti.wIndex = 0x0000;
		cti.wLength = 0x0012;
		cti.data = new unsigned char[cti.wLength];
		if(!control_transfer(&cti)) {
			delete[] cti.data;
			return false;
		}
		delete[] cti.data;
		cti.bmRequestType = 0x02;
		cti.bRequest = 0x01;
		cti.wValue = 0x0000;
		cti.wIndex = 0x0081;
		cti.wLength = 0x0000;
		cti.data = NULL;
		cti.timeout = 0;
		usleep(80000);
		if(!control_transfer(&cti)) return false;
		cti.wIndex = 0x0002;
		if(!control_transfer(&cti)) return false;
		usleep(1300000);
		return true;
	}


	bool ti_usb::reset_pipe_short() {
		ctl_transfer_info cti;

		cti.hdl = ti_usb::hdl;
		cti.bmRequestType = 0x02;
		cti.bRequest = 0x01;
		cti.wValue = 0x0000;
		cti.wIndex = 0x0081;
		cti.wLength = 0x0000;
		cti.data = NULL;
		cti.timeout = 0;
		if(!control_transfer(&cti)) return false;
		cti.wIndex = 0x0002;
		if(!control_transfer(&cti)) return false;
		return true;
	}


	bool ti_usb::control_transfer(ctl_transfer_info *cti) {
		int res;
		res = libusb_control_transfer(cti->hdl, cti->bmRequestType, cti->bRequest,
				cti->wValue, cti->wIndex, cti->data, cti->wLength, cti->timeout);
		return (res == (int)cti->wLength);
	}


	int ti_usb::parse_ls_pkt(unsigned char *pkt, vector<ti_usb::ls_item_info> &res) {
		int pos;

		ls_item_info new_pkt;
		res.push_back(new_pkt);
		ls_item_info *tmpptr = &res[res.size() - 1];
		tmpptr->name = "";
		tmpptr->parent_folder = "";
		if(pkt[0] == 0) {
			pos = 0;
			tmpptr->parent_folder = "";
			pos += 1;
			char tmpchr3[pkt[pos] + 1];
			for(int i = 0; i < (pkt[pos] + 1); i++)
				tmpchr3[i] = pkt[pos + i + 1];
			tmpptr->name.append(tmpchr3);
			pos += (pkt[pos] + 2);
		}
		else {
			pos = 0;
			char tmpchr[pkt[0] + 1];
			for(int i = 0; i < (pkt[0] + 1); i++)
				tmpchr[i] = pkt[i + 1];
			tmpptr->parent_folder.append(tmpchr);
			pos += (pkt[0] + 2);
			if(pkt[pos] == 0) {
				tmpptr->name = "";
				pos += 1;
			} else {
				char tmpchr2[pkt[pos] + 1];
				for(int i = 0; i < (pkt[pos] + 1); i++)
					tmpchr2[i] = pkt[pos + i + 1];
				tmpptr->name.append(tmpchr2);
				pos += (pkt[pos] + 2);
			}
		}
		int num_params = pkt[pos + 1];
		num_params += (pkt[pos] * 0x100);
		pos += 2;
		for(int i = 0; i < num_params; i++) {
			int param_num = (pkt[pos] * 0x100);
			param_num += pkt[pos + 1];
			int valid = pkt[pos + 2];
			pos += 3;
			switch(param_num) {
			case 0x01:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					tmpptr->size = 0;
					for(int i = (len - 1); i >= 0; i--)
						tmpptr->size += (pkt[pos + i] * pow(256, ((len - 1) - i)));
					pos += len;
				} else {
					//Parameter data is not valid
					tmpptr->size = -1;
				}
				break;
			case 0x02:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					tmpptr->type = 0;
					for(int i = (len - 1); i >= 0; i--)
						tmpptr->type += (pkt[pos + i] * pow(256, ((len - 1) - i)));
					pos += len;
					if((tmpptr->type & 0xFF) == 0x1F) {
						string tmp;
						tmp = tmpptr->parent_folder;
						tmpptr->parent_folder = tmpptr->name;
						tmpptr->name = tmp;
					}
				} else {
					//Parameter data is not valid
					tmpptr->type = 0xFFFFFFFF;
				}
				break;
			case 0x03:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					tmpptr->archived = !(pkt[pos] == 0);
					pos += len;
				} else {
					//Parameter data is not valid
					tmpptr->archived = false;
				}
				break;
			case 0x05:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					// Haven't seen this attribute valid yet - don't know what to do
					pos += len;
				} else {
					//Parameter data is not valid
				}
				break;
			case 0x41:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					tmpptr->locked = !(pkt[pos] == 0);
					pos += len;
				} else {
					//Parameter data is not valid
					tmpptr->locked = false;
				}
				break;
			case 0x42:
				if(valid == 0) {
					//Parameter data is valid
					int len = (pkt[pos] * 0x100);
					len += pkt[pos + 1];
					pos += 2;
					// Haven't seen this attribute valid yet - don't know what to do
					pos += len;
				} else {
					//Parameter data is not valid
				}
				break;
			default:
				break;
			}
		}
		return 0;
	}


	bool ti_usb::parameter_request(param_req *req) {
		packet pkt_prsd, pkt_formed;

		if(init_chat() != 0) return false;
		pkt_formed.raw_len = 0x0A;
		pkt_formed.raw_type = 0x04;
		pkt_formed.virt_len = 0x04;
		pkt_formed.virt_type = 0x07;
		pkt_formed.hdl = ti_usb::hdl;
		pkt_formed.payload = new unsigned char[pkt_formed.virt_len];
		pkt_formed.payload[0] = 0x00; pkt_formed.payload[1] = 0x01;
		pkt_formed.payload[2] = 0x00; pkt_formed.payload[3] = req->param_num;
		if(send_pkt(pkt_formed) != 0) {
			delete[] pkt_formed.payload;
			return false;
		}
		delete[] pkt_formed.payload;
		pkt_prsd.hdl = ti_usb::hdl;
		int ret;
		ret = recv_pkt(&pkt_prsd);
		delete[] pkt_prsd.payload;
		if(pkt_prsd.raw_type == 1) {
			bool r;
			libusb_reset_device(ti_usb::hdl);
			r = parameter_request(req);
			return r;
		}
		if(ret != 0) return false;
		if(pkt_prsd.raw_type != 5) return false;
		ret = recv_pkt(&pkt_prsd);
		if(ret != 0) {
			delete[] pkt_prsd.payload;
			return false;
		}
		if(pkt_prsd.virt_type != 0x0008) {
			delete[] pkt_prsd.payload;
			return false;
		}
		if(pkt_prsd.payload[1] != 1) {
			delete[] pkt_prsd.payload;
			return false;
		}
		if(pkt_prsd.payload[3] != req->param_num) {
			delete[] pkt_prsd.payload;
			return false;
		}
		if(pkt_prsd.payload[4] != 0) {
			delete[] pkt_prsd.payload;
			return false;
		}
		for(int i = 0; i < (pkt_prsd.payload_len - 7); i++)
			req->data[i] = pkt_prsd.payload[i + 7];
		req->data_len = (pkt_prsd.payload_len - 7);
		delete[] pkt_prsd.payload;
		return true;
	}


	bool ti_usb::parameter_set(param_req req) {
		packet pkt_prsd, pkt_formed;

		if(init_chat() != 0) return false;
		pkt_formed.raw_len = req.data_len + 10;
		pkt_formed.raw_type = 0x04;
		pkt_formed.virt_len = req.data_len + 4;
		pkt_formed.virt_type = 0x000E;
		pkt_formed.hdl = ti_usb::hdl;
		pkt_formed.payload = new unsigned char[pkt_formed.virt_len + 4];
		pkt_formed.payload[0] = ((req.param_num & 0xFF00) / 0x100);
		pkt_formed.payload[1] = (req.param_num & 0xFF);
		pkt_formed.payload[2] = ((req.data_len & 0xFF00) / 0x100);
		pkt_formed.payload[3] = (req.data_len & 0xFF);
		for(int i = 0; i < req.data_len; i++)
			pkt_formed.payload[i + 4] = req.data[i];
		pkt_formed.payload_len = req.data_len;
		if(send_pkt(pkt_formed) != 0) {
			delete[] pkt_formed.payload;
			return false;
		}
		delete[] pkt_formed.payload;
		pkt_prsd.hdl = ti_usb::hdl;
		int ret;
		ret = recv_pkt(&pkt_prsd);
		delete[] pkt_prsd.payload;
		if(ret != 0) return false;
		if(pkt_prsd.raw_type != 5) return false;
		ret = recv_pkt(&pkt_prsd);
		if(ret != 0) {
			delete[] pkt_prsd.payload;
			return false;
		}
		if(pkt_prsd.virt_type != 0xAA00) {
			delete[] pkt_prsd.payload;
			return false;
		}
		delete[] pkt_prsd.payload;
		return true;
	}


	int ti_usb::send_pkt(packet pk) {
		int len;
		unsigned char *pkt;

		switch(pk.raw_type) {
		case 1:
			len = 9;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = (pk.raw_type & 0xFF);
			pkt[5] = (pk.virt_len & 0xFF000000) / 0x1000000;
			pkt[6] = (pk.virt_len & 0xFF0000) / 0x10000;
			pkt[7] = (pk.virt_len & 0xFF00) / 0x100;
			pkt[8] = (pk.virt_len & 0xFF);
			break;
		case 2:
			len = pk.virt_len + 11;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = (pk.raw_type & 0xFF);
			pkt[5] = (pk.virt_len & 0xFF000000) / 0x1000000;
			pkt[6] = (pk.virt_len & 0xFF0000) / 0x10000;
			pkt[7] = (pk.virt_len & 0xFF00) / 0x100;
			pkt[8] = (pk.virt_len & 0xFF);
			pkt[9] = (pk.virt_type & 0xFF00) / 0x100;
			pkt[10] = (pk.virt_type & 0xFF);
			for(int i = 0; i < pk.virt_len; i++)
				pkt[i + 11] = pk.payload[i];
			break;
		case 3:
			len = pk.raw_len + 5;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = (pk.raw_type & 0xFF);
			for(int i = 0; i < pk.raw_len; i++)
				pkt[i + 5] = pk.payload[i];
			break;
		case 4:
			len = pk.virt_len + 11;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = (pk.raw_type & 0xFF);
			pkt[5] = (pk.virt_len & 0xFF000000) / 0x1000000;
			pkt[6] = (pk.virt_len & 0xFF0000) / 0x10000;
			pkt[7] = (pk.virt_len & 0xFF00) / 0x100;
			pkt[8] = (pk.virt_len & 0xFF);
			pkt[9] = (pk.virt_type & 0xFF00) / 0x100;
			pkt[10] = (pk.virt_type & 0xFF);
			for(int i = 0; i < pk.virt_len; i++)
				pkt[i + 11] = pk.payload[i];
			break;
		case 5:
			len = 7;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = (pk.raw_type & 0xFF);
			pkt[5] = (pk.virt_len & 0xFF00) / 0x100;
			pkt[6] = (pk.virt_len & 0xFF);
			break;
		case 6:
			len = pk.raw_len + 5;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = 0x03;
			pkt[5] = (pk.virt_len & 0xFF000000) / 0x1000000;
			pkt[6] = (pk.virt_len & 0xFF0000) / 0x10000;
			pkt[7] = (pk.virt_len & 0xFF00) / 0x100;
			pkt[8] = (pk.virt_len & 0xFF);
			pkt[9] = (pk.virt_type & 0xFF00) / 0x100;
			pkt[10] = (pk.virt_type & 0xFF);
			for(int i = 0; i < (pk.raw_len - 6); i++)
				pkt[i + 11] = pk.payload[i];
			break;
		case 7:
			len = pk.raw_len + 5;
			pkt = new unsigned char[len];
			pkt[0] = (pk.raw_len & 0xFF000000) / 0x1000000;
			pkt[1] = (pk.raw_len & 0xFF0000) / 0x10000;
			pkt[2] = (pk.raw_len & 0xFF00) / 0x100;
			pkt[3] = (pk.raw_len & 0xFF);
			pkt[4] = 0x04;
			for(int i = 0; i < pk.raw_len; i++)
				pkt[i + 5] = pk.payload[i];
			break;
		default:
			return -1;
		}
		transfer_info tinfo;
		tinfo.direct = out;
		tinfo.hdl = pk.hdl;
		tinfo.data = pkt;
		tinfo.len = &len;
		tinfo.timeout = 0;
		tinfo.actual = 0;
		if(!bulk_transfer(&tinfo)) {
			delete[] pkt;
			return -1;
		}
		delete[] pkt;
		return 0;
	}


	int ti_usb::send_virt_pkt_ack() {
		int ret;
		packet pk;

		pk.raw_len = 0x02;
		pk.raw_type = 0x05;
		pk.virt_len = 0xE000;
		pk.hdl = ti_usb::hdl;
		ret = send_pkt(pk);
		return ret;
	}


	int ti_usb::find_calc() {
		libusb_device **devs;
		struct libusb_device_descriptor desc;
		ssize_t cnt, i;
		int ret;
		string routine;

		cnt = libusb_get_device_list(ti_usb::ctx, &devs);
		if(cnt < 0) return cnt;
		for(i = 0; i < cnt; i++) {
			ret = libusb_get_device_descriptor(devs[i], &desc);
			if(ret < 0) {
				libusb_free_device_list(devs, 1);
				return ret;
			}
			if((desc.idProduct == 0xe004) && (desc.idVendor == 0x0451)) {
				ret = libusb_open(devs[i], &(ti_usb::hdl));
				if(ret < 0) {
					libusb_free_device_list(devs, 1);
					return ret;
				}
				libusb_free_device_list(devs, 1);
				return 0;
			}
		}
		libusb_free_device_list(devs, 1);
		return -1;
	}


	int ti_usb::unload_kern_driver() {
		int r, s;

		r = libusb_kernel_driver_active(ti_usb::hdl, 0);
		if(r <= 0) return r;
		s = 0;
		if(r == 1) s = libusb_detach_kernel_driver(ti_usb::hdl, 0);
		return s;
	}


	int ti_usb::recv_pkt(packet *pkt_prsd) {
		unsigned char sub_pkt[2048], pkt[4096];
		int i, j, len;
		transfer_info tinfo;

		tinfo.direct = in;
		tinfo.hdl = pkt_prsd->hdl;
		tinfo.timeout = pkt_prsd->timeout;
		tinfo.data = sub_pkt;
		len = 256;
		tinfo.len = &len;
		tinfo.actual = 0;
		if(bulk_transfer(&tinfo)) {
			if(tinfo.actual == 0) bulk_transfer(&tinfo);
			// We received a packet - now check to see if it is complete.
			len = tinfo.actual  - 5;
			i = sub_pkt[0] * 0x1000000;
			i += (sub_pkt[1] * 0x10000);
			i += (sub_pkt[2] * 0x100);
			i += sub_pkt[3];
			pkt_prsd->raw_len = i;
			pkt_prsd->raw_type = sub_pkt[4];
			for(j = 0; j < (tinfo.actual - 5); j++) {
				pkt[j] = sub_pkt[j + 5];
			}
			if(pkt_prsd->raw_len == (tinfo.actual - 5)) {
				switch(pkt_prsd->raw_type) {
				case 1: {
					// Should not receive any of these unless calculator has
					// just been reset - then it will pop one out - why?  Don't
					// ask me.
					pkt_prsd->virt_len = -1;
					pkt_prsd->virt_type = -1;
					pkt_prsd->payload = new unsigned char[len];
					for(int i = 0; i < len; i++)
						pkt_prsd->payload[i] = pkt[i];
					pkt_prsd->payload_len = len;
					return 0;
				}
					break;
				case 2: {
					pkt_prsd->virt_len = -1;
					pkt_prsd->virt_type = -1;
					pkt_prsd->payload = new unsigned char[len];
					for(int i = 0; i < len; i++)
						pkt_prsd->payload[i] = pkt[i];
					pkt_prsd->payload_len = len;
					return 0;
				}
					break;
				case 3: {
					// This should not hit - a type 3 is always multi-packet
					// In case it should ever hit we construct a fake pkt
					pkt_prsd->payload = new unsigned char[1];
					return -1;
				}
					break;
				case 4: {
					i = pkt[0] * 0x1000000;
					i += (pkt[1] * 0x10000);
					i += (pkt[2] * 0x100);
					i += pkt[3];
					pkt_prsd->virt_len = i;
					i = pkt[4] * 0x100;
					i += pkt[5];
					pkt_prsd->virt_type = i;
					pkt_prsd->payload = new unsigned char[len - 6];
					for(int i = 0; i < (len - 6); i++)
						pkt_prsd->payload[i] = pkt[i + 6];
					pkt_prsd->payload_len = (len - 6);
					return send_virt_pkt_ack();
				}
					break;
				case 5: {
					pkt_prsd->virt_len = -1;
					pkt_prsd->virt_type = -1;
					pkt_prsd->payload = new unsigned char[len];
					for(int i = 0; i < len; i++)
						pkt_prsd->payload[i] = pkt[i];
					pkt_prsd->payload_len = len;
					return 0;
				}
					break;
				default:
					return -1;
					break;
				}
			}
			tinfo.len = &i;
			tinfo.actual = 0;
			if(bulk_transfer(&tinfo)) {
				switch(pkt_prsd->raw_type) {
				case 1:
					// Should not receive one of these
					break;
				case 2: {
					pkt_prsd->virt_len = -1;
					pkt_prsd->virt_type = -1;
					pkt_prsd->payload = new unsigned char[*tinfo.len];
					for(int i = 0; i < *tinfo.len; i++)
						pkt_prsd->payload[i] = tinfo.data[i];
					pkt_prsd->payload_len = *tinfo.len;
					return 0;
				}
					break; {

					}
				case 3: {
					if(pkt_prsd->multi_in_progress) {
						for(int i = 0; i < pkt_prsd->raw_len; i++)
							pkt_prsd->payload[i + pkt_prsd->payload_len] = tinfo.data[i];
						pkt_prsd->payload_len += pkt_prsd->raw_len;
						if(send_virt_pkt_ack() != 0) {
							return -1;
						}
						return recv_pkt(pkt_prsd);
					}
					pkt_prsd->multi_in_progress = true;
					i = tinfo.data[0] * 0x1000000;
					i += (tinfo.data[1] * 0x10000);
					i += (tinfo.data[2] * 0x100);
					i += tinfo.data[3];
					pkt_prsd->virt_len = i;
					i = tinfo.data[4] * 0x100;
					i += tinfo.data[5];
					pkt_prsd->virt_type = i;
					pkt_prsd->payload = new unsigned char[pkt_prsd->virt_len];
					for(int i = 6; i < pkt_prsd->raw_len; i++)
						pkt_prsd->payload[i - 6] = tinfo.data[i];
					pkt_prsd->payload_len = pkt_prsd->raw_len - 6;
					if(send_virt_pkt_ack() != 0) {
						return -1;
					}
					return recv_pkt(pkt_prsd);
				}
					break;
				case 4: {
					if(!pkt_prsd->multi_in_progress) {
						i = sub_pkt[0] * 0x1000000;
						i += (sub_pkt[1] * 0x10000);
						i += (sub_pkt[2] * 0x100);
						i += sub_pkt[3];
						pkt_prsd->virt_len = i;
						i = sub_pkt[4] * 0x100;
						i += sub_pkt[5];
						pkt_prsd->virt_type = i;
						if(pkt_prsd->virt_type == 0xBB00) {
							int timeout;
							timeout = (sub_pkt[6] * 0x1000000);
							timeout += (sub_pkt[7] * 0x10000);
							timeout += (sub_pkt[8] * 0x100);
							timeout += sub_pkt[9];
							pkt_prsd->timeout = timeout;
							send_virt_pkt_ack();
							if(recv_pkt(pkt_prsd) != 0) {
								return -1;
							}
							if(pkt_prsd->virt_type != 0xAA00) {
								return -1;
							}
							return 0;
						}
						if(pkt_prsd->virt_len > 0) {
							pkt_prsd->payload = new unsigned char[pkt_prsd->virt_len];
							for(int i = 0; i < pkt_prsd->virt_len; i++)
								pkt_prsd->payload[i] = tinfo.data[i + 6];
						}
						pkt_prsd->payload_len = (pkt_prsd->virt_len);
						return(send_virt_pkt_ack());
					}
					for(int i = 0; i < pkt_prsd->raw_len; i++)
						pkt_prsd->payload[i + pkt_prsd->payload_len] = tinfo.data[i];
					pkt_prsd->payload_len += pkt_prsd->raw_len;
					int ret;
					ret = send_virt_pkt_ack();
					if(ret != 0) {
						pkt_prsd->multi_in_progress = false;
						return ret;
					}
					pkt_prsd->multi_in_progress = false;
					return 0;
				}
					break;
				case 5: {
					pkt_prsd->virt_len = -1;
					pkt_prsd->virt_type = -1;
					pkt_prsd->payload = new unsigned char[*tinfo.len];
					for(int i = 0; i < *tinfo.len; i++)
						pkt_prsd->payload[i] = tinfo.data[i];
					pkt_prsd->payload_len = *tinfo.len;
					pkt_prsd->payload = pkt_prsd->payload;
					return 0;
				}
					break;
				default:
					return -1;
					break;
				}
			}
		}
		return -1;
	}


	int ti_usb::init_chat() {
		int ret;
		packet pkt_prsd, pkt_formed;

		pkt_formed.raw_len = 4;
		pkt_formed.raw_type = 1;
		pkt_formed.virt_len = 0x400;
		pkt_formed.hdl = ti_usb::hdl;
		pkt_formed.payload = NULL;
		ret = send_pkt(pkt_formed);
		if(ret != 0) return ret;

		pkt_prsd.hdl = ti_usb::hdl;
		ret = recv_pkt(&pkt_prsd);
		delete[] pkt_prsd.payload;
		if(ret != 0) {
			libusb_reset_device(ti_usb::hdl);
			ret = init_chat();
			return ret;
		}
		if(pkt_prsd.raw_type != 2) return -1;

		pkt_formed.raw_len = 0x10;
		pkt_formed.raw_type = 0x04;
		pkt_formed.virt_len = 0x0A;
		pkt_formed.virt_type = 0x01;
		pkt_formed.payload = new unsigned char[10];
		pkt_formed.payload[0] = 0x00; pkt_formed.payload[1] = 0x03;
		pkt_formed.payload[2] = 0x00; pkt_formed.payload[3] = 0x01;
		pkt_formed.payload[4] = 0x00; pkt_formed.payload[5] = 0x00;
		pkt_formed.payload[6] = 0x00; pkt_formed.payload[7] = 0x00;
		pkt_formed.payload[8] = 0x07; pkt_formed.payload[9] = 0xD0;
		ret = send_pkt(pkt_formed);
		delete[] pkt_formed.payload;
		if(ret != 0) return ret;

		ret = recv_pkt(&pkt_prsd);
		delete[] pkt_prsd.payload;
		if(ret != 0) return ret;
		if(pkt_prsd.raw_type != 5) return -1;

		ret = recv_pkt(&pkt_prsd);
		delete[] pkt_prsd.payload;
		if(ret != 0) return ret;
		if((pkt_prsd.raw_type != 4) || (pkt_prsd.virt_type != 18)) return -1;
		return ret;
	}


	string ti_usb::typeid_to_string(int type_id) {
		if(is_good()) {
			switch(type_id & 0xFF) {
			case 0x00:
				return "EXPR";
				break;
			case 0x04:
				return "FIG";
				break;
			case 0x0C:
				return "STR";
				break;
			case 0x0E:
				return "LIST";
				break;
			case 0x10:
				return "PICT";
				break;
			case 0x12:
				return "PRGM";
				break;
			case 0x13:
				return "FUNCT";
				break;
			case 0x1C:
				return "AppVar";
				break;
			case 0x1F:
				return "Folder";
				break;
			case 0x21:
				return "ASM";
				break;
			case 0x23:
				return "OS";
				break;
			case 0x24:
				return "Flash APP";
				break;
			case 0x25:
				return "Certificate";
				break;
			case 0x3E:
				return "License";
				break;
			default:
				return "Unknown";
				break;
			}
		}
		return "";
	}


	int ti_usb::delete_var(ls_item_info ls_item) {
		if(is_good()) {
			packet pk, pkt_prsd;

			int len_parent_folder = ls_item.parent_folder.length();
			int len_name = ls_item.name.length();
			pk.payload_len = len_parent_folder + len_name + 24;
			pk.payload = new unsigned char[pk.payload_len];
			pk.payload[0] = (len_parent_folder & 0xFF);
			int index = 1;
			if(len_parent_folder > 0) {
				for(int i = index; i < (len_parent_folder + 2); i++)
					pk.payload[i] = ls_item.parent_folder.c_str()[i - 1];
				index += (len_parent_folder + 1);
			}
			pk.payload[index++] = (len_name & 0xFF);
			if(len_name > 0) {
				for(int i = index; i < (index + len_name + 1); i++)
					pk.payload[i] = ls_item.name.c_str()[i - index];
				index += (len_name + 1);
			}
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x02;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x11;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x04;
			pk.payload[index++] = ((ls_item.type & 0xFF000000) / 0x1000000);
			pk.payload[index++] = (((ls_item.type & 0xFF0000) / 0x10000) + 8);
			pk.payload[index++] = ((ls_item.type & 0xFF00) / 0x100);
			pk.payload[index++] = (ls_item.type & 0xFF);
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x13;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			if(ls_item.archived) pk.payload[index++] = 0x01;
			else pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.raw_len = pk.payload_len + 6;
			pk.raw_type = 0x04;
			pk.virt_len = pk.payload_len;
			pk.virt_type = 0x0010;
			pk.hdl = ti_usb::hdl;
			if(!reset_pipe_long()) {
				delete[] pk.payload;
				return -1;
			}
			if(!set_homescreen()) {
				delete[] pk.payload;
				return -1;
			}
			if(send_pkt(pk) != 0) {
				delete[] pk.payload;
				return -1;
			}
			delete[] pk.payload;
			pkt_prsd.hdl = ti_usb::hdl;
			int ret;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return -1;
			if(pkt_prsd.raw_type != 5) return -1;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return -1;
			if(pkt_prsd.virt_type == 0xAA00) {
				reset_pipe_short();
				return 0;
			}
			return -1;
		}
		return -1;
	}


	int ti_usb::get_var_data(ls_item_info ls_item, unsigned char *buf) {
		if(is_good()) {
			packet pk, pkt_prsd;

			int len_parent_folder = ls_item.parent_folder.length();
			int len_name = ls_item.name.length();
			pk.payload_len = len_parent_folder + len_name + 29;
			pk.payload = new unsigned char[pk.payload_len];
			pk.payload[0] = (len_parent_folder & 0xFF);
			int index = 1;
			if(len_parent_folder > 0) {
				for(int i = index; i < (len_parent_folder + 2); i++)
					pk.payload[i] = ls_item.parent_folder.c_str()[i - 1];
				index += (len_parent_folder + 1);
			}
			pk.payload[index++] = (len_name & 0xFF);
			if(len_name > 0) {
				for(int i = index; i < (index + len_name + 1); i++)
					pk.payload[i] = ls_item.name.c_str()[i - index];
				index += (len_name + 1);
			}
			pk.payload[index++] = 0x01;
			pk.payload[index++] = 0xFF;
			pk.payload[index++] = 0xFF;
			pk.payload[index++] = 0xFF;
			pk.payload[index++] = 0xFF;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x03;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x03;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x08;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x41;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x11;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x04;
			pk.payload[index++] = ((ls_item.type & 0xFF000000) / 0x1000000);
			pk.payload[index++] = (((ls_item.type & 0xFF0000) / 0x10000) + 8);
			pk.payload[index++] = ((ls_item.type & 0xFF00) / 0x100);
			pk.payload[index++] = (ls_item.type & 0xFF);
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.raw_len = pk.payload_len + 6;
			pk.raw_type = 0x04;
			pk.virt_len = pk.payload_len;
			pk.virt_type = 0x000C;
			pk.hdl = ti_usb::hdl;
			if(!reset_pipe_long()) {
				delete[] pk.payload;
				return -1;
			}
			if(!set_homescreen()) {
				delete[] pk.payload;
				return -1;
			}
			if(send_pkt(pk) != 0) {
				delete[] pk.payload;
				return -1;
			}
			delete[] pk.payload;
			pkt_prsd.hdl = ti_usb::hdl;
			int ret;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return -1;
			if(pkt_prsd.raw_type != 5) return -1;
			ret = recv_pkt(&pkt_prsd);
			if((pkt_prsd.virt_type != 0x000A) || (ret != 0)) {
				delete[] pkt_prsd.payload;
				return -1;
			}
			int name_length = pkt_prsd.payload[0];
			for(int i = 0; i < (pkt_prsd.payload_len - 1); i++)
				pkt_prsd.payload[i] = pkt_prsd.payload[i + 1];
			pkt_prsd.payload_len--;
			string name;
			name.append((char *)pkt_prsd.payload, name_length);
			for(int i = 0; i < (pkt_prsd.payload_len - (name_length + 1)); i++)
				pkt_prsd.payload[i] = pkt_prsd.payload[i + name_length + 1];
			pkt_prsd.payload_len -= (name_length + 1);
			int num_params = pkt_prsd.payload[1];
			num_params += (pkt_prsd.payload[0] * 0x100);
			for(int i = 0; i < (pkt_prsd.payload_len - 2); i++)
				pkt_prsd.payload[i] = pkt_prsd.payload[i + 2];
			int param_num, is_archived, is_locked, version;
			is_archived = -1;
			is_locked = -1;
			version = -1;
			bool is_ok;
			for(int i = 0; i < num_params; i++) {
				param_num = pkt_prsd.payload[1];
				param_num += (pkt_prsd.payload[0] * 0x100);
				is_ok = (pkt_prsd.payload[2] == 0);
				for(int j = 0; j < 3; j++)
					pkt_prsd.payload[j] = pkt_prsd.payload[j + 3];
				pkt_prsd.payload_len -= 3;
				if(is_ok) {
					switch(param_num) {
					case 3:
						is_archived = pkt_prsd.payload[2];
						for(int j = 0; j < 3; j++)
							pkt_prsd.payload[j] = pkt_prsd.payload[j + 3];
						pkt_prsd.payload_len -= 3;
						break;
					case 8:
						version = 0;
						version += pkt_prsd.payload[5];
						version += (pkt_prsd.payload[4] * 0x100);
						version += (pkt_prsd.payload[3] * 0x10000);
						version += (pkt_prsd.payload[2] * 0x1000000);
						for(int j = 0; j < 6; j++)
							pkt_prsd.payload[j] = pkt_prsd.payload[j + 6];
						pkt_prsd.payload_len -= 6;
						break;
					case 65:
						is_locked = pkt_prsd.payload[2];
						for(int j = 0; j < 3; j++)
							pkt_prsd.payload[j] = pkt_prsd.payload[j + 3];
						pkt_prsd.payload_len -= 3;
						break;
					default:
						break;
					}
				}
			}
			delete[] pkt_prsd.payload;
			ret = recv_pkt(&pkt_prsd);
			if((ret != 0) || (pkt_prsd.virt_len != ls_item.size)) {
				delete[] pkt_prsd.payload;
				return -1;
			}
			for(int i = 0; i < pkt_prsd.payload_len; i++)
				buf[i] = pkt_prsd.payload[i];
			delete[] pkt_prsd.payload;
			reset_pipe_short();
			return 0;
		}
		return -1;
	}


	int ti_usb::put_var(ls_item_info ls_item, unsigned char *data) {
		if(is_good()) {
			packet pk, pkt_prsd;

			if((ls_item.type & 0xFF) == 0x1F) {
				string tmp;
				tmp = ls_item.parent_folder;
				ls_item.parent_folder = ls_item.name;
				ls_item.name = tmp;
			}
			int len_parent_folder = ls_item.parent_folder.length();
			int len_name = ls_item.name.length();
			if(len_parent_folder == 0) pk.payload_len = 1;
			else pk.payload_len = len_parent_folder + 2;
			if(len_name == 0) pk.payload_len += 1;
			else pk.payload_len += (len_name + 2);
			pk.payload_len += 33;
			pk.payload = new unsigned char[pk.payload_len];
			pk.payload[0] = (len_parent_folder & 0xFF);
			int index = 1;
			if(len_parent_folder > 0) {
				for(int i = index; i < (len_parent_folder + 2); i++)
					pk.payload[i] = ls_item.parent_folder.c_str()[i - 1];
				index += (len_parent_folder + 1);
			}
			pk.payload[index++] = (len_name & 0xFF);
			if(len_name > 0) {
				for(int i = index; i < (index + len_name + 1); i++)
					pk.payload[i] = ls_item.name.c_str()[i - index];
				index += (len_name + 1);
			}
			pk.payload[index++] = ((ls_item.size & 0xFF000000) / 0x1000000);
			pk.payload[index++] = ((ls_item.size & 0xFF0000) / 0x10000);
			pk.payload[index++] = ((ls_item.size & 0xFF00) / 0x100);
			pk.payload[index++] = (ls_item.size & 0xFF);
			pk.payload[index++] = 0x01;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x04;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x02;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x04;
			pk.payload[index++] = ((ls_item.type & 0xFF000000) / 0x1000000);
			pk.payload[index++] = ((ls_item.type & 0xFF0000) / 0x10000);
			pk.payload[index++] = ((ls_item.type & 0xFF00) / 0x100);
			pk.payload[index++] = (ls_item.type & 0xFF);
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x03;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			if(ls_item.archived) pk.payload[index++] = 0x01;
			else pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x08;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x04;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x41;
			pk.payload[index++] = 0x00;
			pk.payload[index++] = 0x01;
			if(ls_item.locked) pk.payload[index++] = 0x01;
			else pk.payload[index++] = 0x00;
			pk.raw_len = pk.payload_len + 6;
			pk.raw_type = 0x04;
			pk.virt_len = pk.payload_len;
			pk.virt_type = 0x000B;
			pk.hdl = ti_usb::hdl;
			if(!reset_pipe_long()) {
				delete[] pk.payload;
				return -1;
			}
			if(!set_homescreen()) {
				delete[] pk.payload;
				return -1;
			}
			if(send_pkt(pk) != 0) {
				delete[] pk.payload;
				return -1;
			}
			delete[] pk.payload;
			pkt_prsd.hdl = ti_usb::hdl;
			int ret;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return -1;
			if(pkt_prsd.raw_type != 5) return -1;
			ret = recv_pkt(&pkt_prsd);
			delete[] pkt_prsd.payload;
			if(ret != 0) return -1;
			if(((ls_item.type & 0xFF) == 0x1F) &&
					(pkt_prsd.virt_type == 0xAA00)) {
				pk.raw_len = 0x06;
				pk.raw_type = 0x04;
				pk.virt_len = 0x00;
				pk.virt_type = 0xDD00;
				if(send_pkt(pk) != 0) return -1;
				int ret;
				ret = recv_pkt(&pkt_prsd);
				delete[] pkt_prsd.payload;
				if(ret != 0) return -1;
				if(pkt_prsd.raw_type != 5) return -1;
				return 0;
			}
			if(pkt_prsd.virt_type == 0xAA00) {
				int remaining = ls_item.size;
				int pointer = 0;
				bool first_pkt = true;
				pk.payload = new unsigned char[0x3FF];
				while(remaining > 0) {
					if((remaining < 0x3FA) && first_pkt) {
						pk.raw_len = remaining + 6;
						pk.raw_type = 0x04;
						pk.virt_len = remaining;
						pk.virt_type = 0x000D;
						for(int i = 0; i < remaining; i++)
							pk.payload[i] = data[i + pointer];
						remaining = 0;
						if(send_pkt(pk) != 0) {
							delete[] pk.payload;
							return -1;
						}
						first_pkt = false;
					} else {
						if(first_pkt) {
							pk.raw_len = 0x3FF;
							pk.raw_type = 0x06;
							pk.virt_len = ls_item.size;
							pk.virt_type = 0x000D;
							for(int i = 0; i < 0x3F9; i++)
								pk.payload[i] = data[i + pointer];
							pointer += 0x3F9;
							remaining -= 0x3F9;
							if(send_pkt(pk) != 0) {
								delete[] pk.payload;
								return -1;
							}
							first_pkt = false;
						} else {
							if(remaining > 0x3FF) {
								pk.raw_len = 0x3FF;
								pk.raw_type = 0x07;
								for(int i = 0; i < 0x3FF; i++)
									pk.payload[i] = data[pointer + i];
								pointer += 0x3FF;
								remaining -= 0x3FF;
								if(send_pkt(pk) != 0) {
									delete[] pk.payload;
									return -1;
								}
							} else {
								pk.raw_len = remaining;
								pk.raw_type = 0x07;
								for(int i = 0; i < remaining; i++)
									pk.payload[i] = data[pointer + i];
								remaining = 0;
								if(send_pkt(pk) != 0) {
									delete[] pk.payload;
									return -1;
								}
							}
						}
					}
					int ret;
					ret = recv_pkt(&pkt_prsd);
					delete[] pkt_prsd.payload;
					if(ret != 0) return -1;
					if(pkt_prsd.raw_type != 5) return -1;
				}
				delete[] pk.payload;
				int ret;
				ret = recv_pkt(&pkt_prsd);
				delete[] pkt_prsd.payload;
				if(ret != 0) return -1;
				if(pkt_prsd.virt_type != 0xAA00) return -1;
				pk.raw_len = 0x06;
				pk.raw_type = 0x04;
				pk.virt_len = 0x00;
				pk.virt_type = 0xDD00;
				if(send_pkt(pk) != 0) {
					return -1;
				}
				ret = recv_pkt(&pkt_prsd);
				delete[] pkt_prsd.payload;
				if(ret != 0) return -1;
				if(pkt_prsd.raw_type != 0x05) return -1;
				reset_pipe_short();
				return 0;
			}
			return -1;
		}
		return -1;
	}


	int ti_usb::var_save(ls_item_info ls_item, unsigned char *buf, string path) {
		if(is_good()) {
			if((ls_item.type & 0xFF) == 0x24) {
				ti_usb::flash_head fh;
				fh.name = ls_item.name;
				fh.version[0] = 0x00;
				fh.version[1] = 0x00;
				ti_usb::date_time dt;
				bool ret;
				ret = get_sys_time(&dt);
				if(ret) {
					fh.bcd_date[0] = (dt.day & 0xFF);
					fh.bcd_date[1] = (dt.month & 0xFF);
					int year_h, year_l;
					year_h = (((dt.year / 100) * 16) / 10);
					year_l = (((dt.year - ((dt.year / 100) * 100)) * 16) / 10);
					fh.bcd_date[2] = (year_h & 0xFF);
					fh.bcd_date[3] = (year_l & 0xFF);
				} else {
					fh.bcd_date[0] = 0;
					fh.bcd_date[1] = 0;
					fh.bcd_date[2] = 0;
					fh.bcd_date[3] = 0;
				}
				fh.dev_type = 0x88;
				fh.data_type = (ls_item.type & 0xFF);
				fh.hdw_id = 0x0;
				fh.data_size = ls_item.size;
				fh.data = buf;
				string sig = "**TIFL**";
				unsigned char *filebuf = new unsigned char[0x4E + ls_item.size];
				int index = 0;
				for(int i = 0; i < (int)sig.length(); i++)
					filebuf[index++] = sig.c_str()[i];
				filebuf[index++] = fh.version[0];
				filebuf[index++] = fh.version[1];
				filebuf[index++] = 0;
				filebuf[index++] = 0;
				filebuf[index++] = fh.bcd_date[0];
				filebuf[index++] = fh.bcd_date[1];
				filebuf[index++] = fh.bcd_date[2];
				filebuf[index++] = fh.bcd_date[3];
				filebuf[index++] = (ls_item.name.length() & 0xFF);
				for(int i = 0; i < (int)ls_item.name.length(); i++)
					filebuf[index++] = ls_item.name.c_str()[i];
				if(ls_item.name.length() < 8)
					for(int i = 0; i < (8 - (int)ls_item.name.length()); i++)
						filebuf[index++] = 0;
				for(int i = 0; i < 23; i++)
					filebuf[index++] = 0;
				filebuf[index++] = (fh.dev_type & 0xFF);
				filebuf[index++] = (fh.data_type & 0xFF);
				for(int i = 0; i < 22; i++)
					filebuf[index++] = 0;
				filebuf[index++] = ((fh.hdw_id & 0xFF00) / 0x100);
				filebuf[index++] = (fh.hdw_id & 0xFF);
				filebuf[index++] = (fh.data_size & 0xFF);
				filebuf[index++] = ((fh.data_size & 0xFF00) / 0x100);
				filebuf[index++] = ((fh.data_size & 0xFF0000) / 0x10000);
				filebuf[index++] = ((fh.data_size & 0xFF000000) / 0x1000000);
				for(int i = 0; i < fh.data_size; i++)
					filebuf[index++] = buf[i];
				string filename;
				filename = path + "main." + fh.name + ".89k";
				if(!write_file(filename, filebuf, 0x4E + fh.data_size)) {
					delete[] filebuf;
					return -1;
				}
				delete[] filebuf;
				return 0;
			} else {
				int type = (ls_item.type & 0xFF);
				if((type == 0x00) || (type == 0x04) || (type == 0x0C) ||
						(type == 0x0E) || (type == 0x10) || (type == 0x12) ||
						(type == 0x13)) {
					int filesize = ls_item.size + 0x58;
					string sig = "**TI92P*";
					unsigned char *filebuf = new unsigned char[filesize];
					int index = 0;
					for(int i = 0; i < 8; i++)
						filebuf[index++] = sig.c_str()[i];
					filebuf[index++] = 0x01;
					filebuf[index++] = 0x00;
					if(ls_item.parent_folder.length() > 0)
						for(int i = 0; i < (int)ls_item.parent_folder.length(); i++)
							filebuf[index++] = ls_item.parent_folder.c_str()[i];
					if(ls_item.parent_folder.length() < 8)
						for(int i = 0; i < (8 - (int)ls_item.parent_folder.length()); i++)
							filebuf[index++] = 0;
					for(int i = 0; i < 40; i++)
						filebuf[index++] = 0;
					filebuf[index++] = 0x01;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x52;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					for(int i = 0; i < (int)ls_item.name.length(); i++)
						filebuf[index++] = ls_item.name.c_str()[i];
					if(ls_item.name.length() < 8)
						for(int i = 0; i < (8 - (int)ls_item.name.length()); i++)
							filebuf[index++] = 0;
					filebuf[index++] = (ls_item.type & 0xFF);
					int arch_lock = 0;
					if(ls_item.archived) arch_lock += 2;
					if(ls_item.locked) arch_lock += 1;
					filebuf[index++] = (arch_lock & 0xFF);
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					filebuf[index++] = (filesize & 0xFF);
					filebuf[index++] = ((filesize & 0xFF00) / 0x100);
					filebuf[index++] = ((filesize & 0xFF0000) / 0x10000);
					filebuf[index++] = ((filesize & 0xFF000000) / 0x1000000);
					filebuf[index++] = 0xA5;
					filebuf[index++] = 0x5A;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					filebuf[index++] = 0x00;
					unsigned int cs = 0;
					for(int i = 0; i < (int)ls_item.size; i++) {
						filebuf[index++] = buf[i];
						cs += buf[i];
					}
					filebuf[index++] = (cs & 0xFF);
					filebuf[index++] = ((cs & 0xFF00) / 0x100);
					string filename;
					if(ls_item.parent_folder == "") ls_item.parent_folder = "main";
					filename = path + ls_item.parent_folder + "." +
							ls_item.name + "." + typeid_to_filext(ls_item.type & 0xFF);
					if(!write_file(filename, filebuf, filesize)) {
						delete[] filebuf;
						return -1;
					}
					delete[] filebuf;
				}
				return 0;
			}
			return -1;
		}
		return -1;
	}


	bool ti_usb::write_file(string filename, unsigned char *buf, int size) {
		struct stat st;
		bool exist, is_dir;

		exist = (stat(filename.c_str(), &st) == 0);
		is_dir = (S_ISDIR(st.st_mode));
		if(!exist && !is_dir) {
			ofstream tifile(filename.c_str(), ios::out | ios::binary);
			if(tifile.is_open()) {
				tifile.write((char *)buf, size);
				tifile.close();
				return true;
			}
			return false;
		}
		return false;
	}


	string ti_usb::typeid_to_filext(int type_id) {
		switch(type_id) {
		case 0x00:
			return "89e";
			break;
		case 0x04:
			return "89a";
			break;
		case 0x0C:
			return "89s";
			break;
		case 0x0E:
			return "89l";
			break;
		case 0x10:
			return "89i";
			break;
		case 0x12:
			return "89p";
			break;
		case 0x13:
			return "89f";
			break;
		default:
			return "";
			break;
		}
	}


	int ti_usb::get_sys_directory(string dir, vector<std::string> &files) {
	    DIR *dp;
	    struct dirent *dirp;
	    if((dp  = opendir(dir.c_str())) == NULL) {
	        return errno;
	    }

	    while ((dirp = readdir(dp)) != NULL) {
	        files.push_back(string(dirp->d_name));
	    }
	    closedir(dp);
	    return 0;
	}


